% DTD & TDOA localization using closed-form solution and MLE.
%
% This code generates Figs. 7(by setting K=2) and 9(by setting K=8) 
% in the reference paper.
%
% Note: Number of ensemble runs (NumEns) is set to 50 only to reduce 
% run time.
%
% Reference:
% Y. Zhang and K. C. Ho, "Localization by signals of opportunity in the
% absence of transmitter position," IEEE Trans. Signal Process., vol. 72,
% pp. 4602-4617, 2022.
%
% Yang Zhang, K. C. Ho	10-18-2022
%
%       Copyright (C) 2022
%       Computational Intelligence Signal Processing Laboratory
%       University of Missouri
%       Columbia, MO 65211, USA
%       hod@missouri.edu
%

clear;
warning('off');

K=2;        % dimension
% K=3;        % dimension

rng('default');

NumEns=50;      % ensemble runs
% NumEns=5000;    % ensemble runs
M=5;        % number of receivers
NoG = 10;   % number of geometries
RgnLgth=2000;   %region length
PlotGeo=0;
CorNse=0;   % correlated measurement noise

for dd  = 1:NoG
    ss(:,:,dd)=RgnLgth*(rand(K,M)-0.5);
    uu(:,dd)=RgnLgth*(rand(K,1)-0.5);
    tt(:,dd)=RgnLgth*(rand(K,1)-0.5);
end

Hr=[-ones(M-1,1),eye(M-1)];

% -- loop over geometries --
for dd = 1:NoG
    S = ss(:,:,dd);
    uo = uu(:,dd);
    to = tt(:,dd);
    
    % -- show geometries --
    if (PlotGeo)
        figure(11);
        if (K==2)
            plot(uo(1,1),uo(2,1),'+k'); hold on;
            plot(to(1,1),to(2,1),'^k'); hold on;
            plot(S(1,:),S(2,:),'ok'); hold off;
            xlabel('x (m)');ylabel('y (m)');
        else
            plot3(uo(1,1),uo(2,1),uo(3,1),'+k'); hold on;
            plot3(to(1,1),to(2,1),to(3,1),'^k'); hold on;
            plot3(S(1,:),S(2,:),S(3,:),'ok'); hold off;
            xlabel('x (m)');ylabel('y (m)'); zlabel('z (m)');
        end;
        grid on;
        lgd=legend('Object','Transmitter','Receiver');
    end;
    
    % -- true DTDs and TDOAs --
    dID = sqrt(sum((repmat(uo,1,M)-S).^2))'+norm(uo-to);
    dD = sqrt(sum((repmat(to,1,M)-S).^2))';
    do = dID - dD;     % differential TDOA measurements
    ro= Hr*dID;         % Indirect-path TDOA measurements
    mo = [do;ro];  % Integrated measurements
    
    nsePwrAlldB=[-20:5:20];     % measurement noise power
    if (K==3) nsePwrAlldB=nsePwrAlldB-20; end;
    
    thetao=[uo; to];    % true object and transmitter positions
    
    % -- loop over noise power --
    for nsePwrIdx=1:length(nsePwrAlldB)
        rng(0);
        fprintf('Geo:%d, 10log(nsePwr):%d\n',dd,nsePwrAlldB(nsePwrIdx));
        nsePwr=10^(nsePwrAlldB(nsePwrIdx)/10);
        m_2=1/(2*M)*sum(dID.^2+dD.^2);
        
        sigmar_2=nsePwr*(dID.^2/m_2);
        sigmad_2=nsePwr*(dD.^2/m_2);
        QdID=diag(sigmar_2)/2;
        QdD=diag(sigmad_2)/2;
        
        % -- correlated noise --
        if (CorNse)
            CorCoef=0.5;
            for i=1:M-1
                for j=i+1:M
                    QdID(i,j)=sqrt(sigmar_2(i))*sqrt(sigmar_2(j))*(norm(uo-to)^2)/((norm(uo-to)+norm(uo-S(:,i)))*(norm(uo-to)+norm(uo-S(:,j))));
                    QdD(i,j)=sqrt(sigmad_2(i))*sqrt(sigmad_2(j))*CorCoef;
                end
            end
        end;
        
        % -- measurement covariance matrix --
        QdID=(QdID+QdID');   % Indirect-path Covariance
        QdD=(QdD+QdD');   % Direct-path Covariance
        Qd=QdID+QdD;       % Covariance of differential TDOA measurements
        Qm = [Qd QdID*Hr';
            Hr*QdID' Hr*QdID*Hr'];    % Covariance of integrated measurements
        
        % -- CRLB --
        CRLB_m=DTDTDOALoc_CRLB(uo,to,S,Qm);    % CRLB using hybrid measurements
        CRLBGeo(dd,nsePwrIdx)=trace(CRLB_m);
        CRLBuGeo(dd,nsePwrIdx)=trace(CRLB_m(1:K,1:K));
        CRLBtGeo(dd,nsePwrIdx)=trace(CRLB_m(K+1:2*K,K+1:2*K));
        
        % -- initializ record --
        SimuMseuMLE=0;
        SimuMseuIni=0;
        SimuMseuMult=0;
        SimuMseuRfn=0;
        
        SimuBiasuMult=0;
        SimuBiasuRfn=0;
        SimuBiasuMLE=0;
        
        % -- ensemble runs --
        for ii=1:NumEns
            
            % -- measurements --
            noise=sqrtm(Qm)*randn(2*M-1,1);
            m=mo+noise;  % noisy measurements
            
            % -- proposed closed-form estimator --
            [theta]=DTDTDOALoc(m,S,Qm); % multistage estimator
            
            % -- proposed refinment solution --
            [thetaHat, ite_Rfn]=DTDTDOALocRfn(m,S,Qm,theta); % refinement method
            
            % -- MLE --
            [theta_MLE, ite_MLE]=DTDTDOALoc_MLE(m,S,Qm,theta); % MLE
            
            % -- collecting statistics --
            SimuMseuMult=SimuMseuMult+norm((theta(1:K)-uo),2).^2;
            SimuMseuRfn=SimuMseuRfn+norm((thetaHat(1:K)-uo),2).^2;
            SimuMseuMLE = SimuMseuMLE + norm((theta_MLE(1:K)-uo),2).^2;
            
            SimuBiasuMult=SimuBiasuMult+(theta(1:K)-uo);
            SimuBiasuRfn=SimuBiasuRfn+(thetaHat(1:K)-uo);
            SimuBiasuMLE = SimuBiasuMLE + (theta_MLE(1:K)-uo);
            
        end
        
        MseuMultGeo(dd,nsePwrIdx)=SimuMseuMult/NumEns;
        MseuRfnGeo(dd,nsePwrIdx)=SimuMseuRfn/NumEns;
        MseuMLEGeo(dd,nsePwrIdx)=SimuMseuMLE/NumEns;
        
        BiasuMultGeo(dd,nsePwrIdx)=sum((SimuBiasuMult/NumEns).^2);
        BiasuRfnGeo(dd,nsePwrIdx)=sum((SimuBiasuRfn/NumEns).^2);
        BiasuMLEGeo(dd,nsePwrIdx)=sum((SimuBiasuMLE/NumEns).^2);
        
    end
    
    % -- plot individual geometry results --
    figure(1);
    subplot(2,5,dd);
    plot(nsePwrAlldB,10*log10(CRLBuGeo(dd,:)),'-r');
    hold on;
    plot(nsePwrAlldB,10*log10(MseuMultGeo(dd,:)),'ob','MarkerSize',8);
    hold on
    plot(nsePwrAlldB,10*log10(MseuRfnGeo(dd,:)),'sm','MarkerSize',8);
    hold on
    plot(nsePwrAlldB,10*log10(MseuMLEGeo(dd,:)),'+k','MarkerSize',8);
    hold off;
    grid on; xlabel('10 log_1_0(\sigma^2(m^2))'); ylabel('10 log_1_0(MSE(u)(m^2))');
    title(['Geo. ',num2str(dd)]);
    if (K==2)
        axis([-20 20 -20 50]);
    else
        axis([-40 0 -40 40]);
    end;

end

% -- average results over geometries --
CRLBu = mean(CRLBuGeo,1);
MseuMult = mean(MseuMultGeo,1);
MseuRfn = mean(MseuRfnGeo,1);
MseuMLE = mean(MseuMLEGeo,1);

BiasuMult = mean(BiasuMultGeo,1);
BiasuRfn = mean(BiasuRfnGeo,1);
BiasuMLE = mean(BiasuMLEGeo,1);

% -- show results --
figure(2);  % MSE
clf;
plot(nsePwrAlldB,10*log10(CRLBu),'-r','LineWidth', 1.25);
hold on
plot(nsePwrAlldB,10*log10(MseuMult),'ob','MarkerSize',8,'LineWidth', 1.25);
hold on
plot(nsePwrAlldB,10*log10(MseuRfn),'sm','MarkerSize',8,'LineWidth', 1.25);
hold on
plot(nsePwrAlldB,10*log10(MseuMLE),'+k','MarkerSize',8,'LineWidth', 1.25);
hold off;
grid on; xlabel('10 log_1_0(\sigma^2(m^2))'); ylabel('10 log_1_0(MSE(u(m)))');
legend('CRLB','Algebraic Solution, No Refinement','Algebraic Solution, With Refinement','MLE by Gauss-Newton','location','northwest');
if (K==2)
    axis([-20 20 -20 50]);
else
    axis([-40 0 -40 40]);
end;

figure(3);  % Bias
clf;
plot(nsePwrAlldB,10*log10(BiasuMult),'ob','MarkerSize',8,'LineWidth', 1.25);
hold on
plot(nsePwrAlldB,10*log10(BiasuRfn),'sm','MarkerSize',8,'LineWidth', 1.25);
hold on
plot(nsePwrAlldB,10*log10(BiasuMLE),'+k','MarkerSize',8,'LineWidth', 1.25);
hold off;
grid on; xlabel('10 log10(nsePwr)'); ylabel('20 log_1_0(Bias(u(m)))');
legend('Algebraic Solution, No Refinement','Algebraic Solution, with Refinement','MLE by Gauss-Newton','location','northwest');
if (K==2)
    axis([-20 20 -40 40]);
else
    axis([-40 0 -80 40]);
end;



